<?php
namespace Fiscal\Logic\NotaFiscal;

use \Fiscal\Entities\Itreplicdoc;
use \Fiscal\Entities\Notafiscal;
use \Fiscal\Entities\Replicdoc;
use \Fiscal\Entities\Trabreplicdoc;
use \Geral\Entities\NovoCodigo;
use \Geral\Util\Ambiente;
use \Setforms2\Runtime\Helper\Procedure;
use \Setforms2\Runtime\Helper\Procedure\Param;


/**
 * Created by JetBrains PhpStorm.
 * User: paulopereira
 * Date: 29/10/12
 * Time: 15:15
 * To change this template use File | Settings | File Templates.
 */
class ReplicacaoAutomatica
{

    /** @var $notafiscal Notafiscal*/
    private $notaFiscal;
    /** @var $nrorg int */
    private $nrorg;


    /**
     * @param \Fiscal\Entities\Notafiscal $notaFiscal
     */
    public function __construct(Notafiscal $notaFiscal)
    {
        $this->notaFiscal = $notaFiscal;
        $this->nrorg      = \Geral\Util\Ambiente::getInstancia()->getNrorg();
    }

    /**
     * Verifica se a nota fiscal deve ou nao ser replicada.
     *
     * @static
     * @return bool
     */
    public static function permiteReplicacaoAutomatica()
    {
        /** @var $paramgeral \Geral\Entities\Paramgeral */
        $paramgeral = \Geral\Entities\Paramgeral::findOneByNrorg(Ambiente::getNrorg());
        return $paramgeral->getIdempdestrepaut() === \Geral\Entities\Paramgeral::IDEMPDESTREPAUT_SIM;
    }

    /**
     * Metodo principal da classe. Responsavel por executar o processo completo de replicacao de notas fiscais. Ele
     * fara chamada aos metodos de replicacao dos documentos(que fara a preparacao dos objetos) e das procedures de
     * replicacao de documentos.
     */
    public function geraReplicacaoAutomatica()
    {
        try {
            //if ($this->existemDocumentosReplicados()) {
            $replicacaoDocumentos = Replicdoc::findByCdfilialAndCdtipooperAndIdentrsaidopAndIdreplautonfAndNrorg (
                $this->notaFiscal->getFilial()->getCdfilial(),
                $this->notaFiscal->getEntrsaidop()->getCdtipooper(),
                \Geral\Entities\Tipooperac::OPERACAO_ENTRADA,
                Replicdoc::PERMITE_REPLICA_AUTOMATICA_NF,
                $this->nrorg
            );
            $this->replicaDocumentos($replicacaoDocumentos);
            $this->executaProcedureReplicacaoDocumentos($replicacaoDocumentos);
            // }
        } catch (\Exception $e) {
            throw $e;
        }
    }

    /**
     * Prepara e executa a procedure de REPLICACAO_DOCUMENTOS.
     *
     * @param array $documentosReplicas
     */
    protected function executaProcedureReplicacaoDocumentos(array $documentosReplicas)
    {
        try {
            /** @var $novoCodigoNF Novocodigo */
            /** @var $novoCodigoEst Novocodigo */
            $intervalo     = count($documentosReplicas);
            $novoCodigoNF  = null;
            $novoCodigoEst = null;
            $procedure     = new Procedure("REPLICACAO_DOCUMENTOS");

            $this->geraNovoCodigo($intervalo, $novoCodigoNF, $novoCodigoEst);
            $this->setParametrosReplicacaoDocumentos($procedure, $novoCodigoNF, $novoCodigoEst);
            $procedure->call();
        } catch (\Exception $e) {
            throw $e;
        }
    }

    /**
     * Gera um novo valor válido para as chaves de Notafiscal e Lancamento de estoque.
     * Obs.: Apesar de nao ser necessario, na assinatura do metodo foi colocado o indicador de variaveis de referencia
     * para os objetos de Novocodigo somente para que a IDE reconheca e nao "reclame" de variaveis nao usadas.
     *
     * @param $intervalo
     * @param \Geral\Entities\NovoCodigo $novoCodigoNF
     * @param \Geral\Entities\NovoCodigo $novoCodigoEst
     */
    protected function geraNovoCodigo($intervalo, Novocodigo &$novoCodigoNF = null, Novocodigo &$novoCodigoEst = null)
    {
        /** @var $cdcontadorNF string Contador de nota fiscal na tabela de novoCodigo.*/
        /** @var $cdcontadorLancest string Contador de lancamento de estoque na tabela de novoCodigo.*/
        $cdcontadorNF      = Notafiscal::CDCONTADOR_NOTAFISCAL . $this->notaFiscal->getFilial()->getCdfilial();
        $cdcontadorLancest = Notafiscal::CDCONTADOR_LANC_ESTOQUE . $this->notaFiscal->getFilial()->getCdfilial();

        try {
            $novoCodigoNF  = NovoCodigo::findOneByCdcontadorAndNrorg($cdcontadorNF, $this->nrorg);
            $novoCodigoEst = NovoCodigo::findOneByCdcontadorAndNrorg($cdcontadorLancest, $this->nrorg);

            \Geral\Util\NovoCodigo::geraCodigo($this->nrorg, $cdcontadorNF, $intervalo);
            \Geral\Util\NovoCodigo::geraCodigo($this->nrorg, $cdcontadorLancest, $intervalo);
        } catch (\Exception $e) {
            throw $e;
        }
    }

    /**
     * Adiciona os parametros da procedure de REPLICACAO_DOCUMENTO.
     *
     * @param \Setforms2\Runtime\Helper\Procedure $procedure
     * @param \Geral\Entities\NovoCodigo $novoCodigoNF
     * @param \Geral\Entities\NovoCodigo $novoCodigoEst
     */
    protected function setParametrosReplicacaoDocumentos(Procedure $procedure, Novocodigo $novoCodigoNF,
                                                         Novocodigo $novoCodigoEst)
    {
        $procedure->addParam(
            new Param (
                "P_CDFILIAL",
                Param::PARAM_INPUT,
                ((int)$novoCodigoNF->getNrsequencial()) + 1,
                Param::PARAM_TYPE_INT
            )
        );
        $procedure->addParam(
            new Param (
                "P_NRLANCTONF",
                Param::PARAM_INPUT,
                $this->notaFiscal->getNrlanctonf(),
                Param::PARAM_TYPE_STR
            )
        );
        $procedure->addParam(
            new Param (
                "P_CDFILIPROC",
                Param::PARAM_INPUT,
                $this->notaFiscal->getFilial()->getCdfilial(),
                Param::PARAM_TYPE_STR
            )
        );
        $procedure->addParam(
            new Param (
                "P_PRXNOTAFISC",
                Param::PARAM_INPUT,
                \Geral\Util\NovoCodigo::getCodigoTamanhoCampo($novoCodigoNF->getNrsequencial(), 10),
                Param::PARAM_TYPE_STR
            )
        );
        $procedure->addParam(
            new Param (
                "P_PRXLANCETSQ",
                Param::PARAM_INPUT,
                \Geral\Util\NovoCodigo::getCodigoTamanhoCampo($novoCodigoEst->getNrsequencial(), 10),
                Param::PARAM_TYPE_STR
            )
        );
    }

    /**
     * Percorre o array de $documentosReplicas e executa a replicacao destes dos itens da replicacao nos itens de
     * trabalho (usados pelas procedures de replicacao de notas fiscais).
     *
     * @param $documentosReplicas
     */
    protected function replicaDocumentos($documentosReplicas)
    {
        /** @var $replicdoc Replicdoc */
        foreach ($documentosReplicas as $replicdoc) {
            $cdfilial              = $replicdoc->getFilial()->getCdfilial();
            $nrreplicdoc           = $replicdoc->getNrreplicdoc();
            $itensDocumentoReplica = Itreplicdoc::findCdfilialAndNrreplicdocAndNrorg (
                $cdfilial, $nrreplicdoc, $this->nrorg
            );
            if ($itensDocumentoReplica) {
                $this->geraReplicasDeTrabalho($itensDocumentoReplica);
            } else {
                $this->geraReplicaDeTrabalhoSemItens();
            }
        }
    }

    /**
     * Executa a replicacao dos objetos de Itreplicdoc passados por parametros em instancias de Trabreplicdoc.
     *
     * @param array $itensDocumentoReplica
     */
    protected function geraReplicasDeTrabalho(array $itensDocumentoReplica)
    {
        /** @var $itreplicdoc Itreplicdoc */
        foreach ($itensDocumentoReplica as $itreplicdoc) {
            $replicaDeTrabalho = new Trabreplicdoc();
            $replicaDeTrabalho->setCdfilialtrd($this->notaFiscal->getFilial()->getCdfilial());
            $replicaDeTrabalho->setNrlanctonftrd($this->notaFiscal->getNrlanctonf());
            $replicaDeTrabalho->setCdfiliorigtrd($itreplicdoc->getFiliorigrd()->getCdfilial());
            $replicaDeTrabalho->setCdfilidesttrd($itreplicdoc->getFilidestrd()->getCdfilial());
            $replicaDeTrabalho->setIdentrsaidtrd($itreplicdoc->getTipooper()->getIdentrsaidop());
            $replicaDeTrabalho->setCdtipoopertrd($itreplicdoc->getTipooper()->getCdtipooper());
            $replicaDeTrabalho->setCdserietrd($itreplicdoc->getSerie()->getCdserie());
            $replicaDeTrabalho->setCdalmoxartrd($this->getAlmoxarifadoPadrao (
                $itreplicdoc->getTipooper()->getIdentrsaidop())
            );
            $replicaDeTrabalho->setCdlocestqtrd($this->getEstoquePadrao (
                $itreplicdoc->getTipooper()->getIdentrsaidop())
            );
            $replicaDeTrabalho->setNrorg($this->nrorg);
            $replicaDeTrabalho->persist();
        }
    }

    /**
     * Executa a replicacao da nota fiscal para casos onde nao haja registros na tabela de item ITREPLICDOC.
     */
    protected function geraReplicaDeTrabalhoSemItens()
    {
        $replicaDeTrabalho = new Trabreplicdoc();
        $replicaDeTrabalho->setCdfilialtrd($this->notaFiscal->getFilial()->getCdfilial());
        $replicaDeTrabalho->setNrlanctonftrd($this->notaFiscal->getNrlanctonf());
        $replicaDeTrabalho->setCdfiliorigtrd($this->notaFiscal->getFiliemit()->getCdfilial());
        $replicaDeTrabalho->setCdfilidesttrd($this->notaFiscal->getFilirece()->getCdfilial());
        $replicaDeTrabalho->setIdentrsaidtrd(\Geral\Entities\Tipooperac::OPERACAO_SAIDA);
        $replicaDeTrabalho->setCdtipoopertrd($this->notaFiscal->getEntrsaidop()->getCdtipooper());
        $replicaDeTrabalho->setCdserietrd($this->notaFiscal->getSerie()->getCdserie());
        $replicaDeTrabalho->setCdalmoxartrd($this->getAlmoxarifadoPadrao(\Geral\Entities\Tipooperac::OPERACAO_SAIDA));
        $replicaDeTrabalho->setCdlocestqtrd($this->getEstoquePadrao(\Geral\Entities\Tipooperac::OPERACAO_SAIDA));
        $replicaDeTrabalho->persist();
    }

    /**
     * Verifica se a nota fiscal em processamento já possui documentos replicados.
     *
     * @return bool
     */
    public function existemDocumentosReplicados()
    {
        $iddocreplic = $this->notaFiscal->getFilial()->getCdfilial().$this->notaFiscal->getNrlanctonf();
        $nf = Notafiscal::FindByIddocreplicAndNrorg($iddocreplic, $this->nrorg);
        return (bool)$nf;
    }

    /**
     * Retorna a o almoxarifado padrao da filial. Caso esta nao possua, retornara a string de identificacao de filial
     * sem almoxarifado padrao.
     *
     * @param $tipoOperacao
     * @return \Geral\Entities\Almoxarifado|string
     */
    private function getAlmoxarifadoPadrao($tipoOperacao)
    {
        /** @var $filial \Geral\Entities\Filial */
        $filial = null;

        if ($tipoOperacao === \Geral\Entities\Tipooperac::OPERACAO_ENTRADA) {
            $filial = $this->notaFiscal->getFilirece();
        } else {
            $filial = $this->notaFiscal->getFiliemit();
        }

        return $filial->getAlmoxarifadoPadrao();
    }

    /**
     * Retorna o estoque padrao da filial. Caso esta nao possua, retornara a string de identificacao de filial sem
     * estoque padrao.
     *
     * @param $tipoOperacao
     * @return \Geral\Entities\Localestoque|string
     */
    private function getEstoquePadrao($tipoOperacao)
    {
        /** @var $filial \Geral\Entities\Filial */
        $filial = null;

        if ($tipoOperacao === \Geral\Entities\Tipooperac::OPERACAO_ENTRADA) {
            $filial = $this->notaFiscal->getFilirece();
        } else {
            $filial = $this->notaFiscal->getFiliemit();
        }

        return $filial->getEstoquePadrao();
    }


}